#include "VertexBasedKnife.h"
#include "MQ3DLib.h"
#include <cstdlib>
#include <cfloat>
#include <cmath>

extern HINSTANCE hInstance;

float GetDlgItemFloat(HWND hDlg, int nIDDlgItem);
BOOL SetDlgItemFloat(HWND hDlg, int nIDDlgItem, float value);
void SetAllScaleEdit(HWND hDlg, float value);
void NormalizeEditBox(HWND hDlg, int nIDDlgItem);

LRESULT CALLBACK MainDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);

struct MQRGBAColor
{
	float r,g,b,a;
	MQRGBAColor() {}
	MQRGBAColor(float _r,float _g,float _b,float _a) {r=_r; g=_g; b=_b; a=_a;}
	MQRGBAColor(const MQRGBAColor& color) {r=color.r; g=color.g; b=color.b; a=color.a;}
	MQRGBAColor(DWORD color) {
		r=(float)( color&0x000000FF)/255;
		g=(float)((color&0x0000FF00)>>8)/255;
		b=(float)((color&0x00FF0000)>>16)/255;
		a=(float)((color&0xFF000000)>>24)/255;
	}

	friend inline MQRGBAColor operator + (const MQRGBAColor& c1, const MQRGBAColor& c2)
		{return MQRGBAColor(c1.r+c2.r, c1.g+c2.g, c1.b+c2.b, c1.a+c2.a);}
	friend inline MQRGBAColor operator * (const MQRGBAColor& c, const float& s)
		{return MQRGBAColor(c.r*s, c.g*s, c.b*s, c.a*s);}

	DWORD GetRGBA() const {
		return (BYTE)(r*255.0f+0.5f) | 
			  ((BYTE)(g*255.0f+0.5f) << 8) |
			  ((BYTE)(b*255.0f+0.5f) << 16) |
			  ((BYTE)(a*255.0f+0.5f) << 24);
		}
};


//---------------------------------------------------------------------------
//  IsCrossing2D
//---------------------------------------------------------------------------
static bool IsCrossing2D(
	const MQPoint& s0, const MQPoint& s1,
	const MQPoint& t0, const MQPoint& t1,
	float *outs, float *outt)
{
	float f,s,t;

	f = (s1.x-s0.x)*(t1.y-t0.y) - (s1.y-s0.y)*(t1.x-t0.x);
	if(f == 0) return false;
	s = ((s0.y-t0.y)*(t1.x-t0.x) - (s0.x-t0.x)*(t1.y-t0.y)) / f;
	t = ((s0.y-t0.y)*(s1.x-s0.x) - (s0.x-t0.x)*(s1.y-s0.y)) / f;
	if(0<=t && t<1 && 0<=s && s<1){
		if(outs != NULL) *outs = s;
		if(outt != NULL) *outt = t;
		return true;
	}

	return false;
}

VertexBasedKnife *VertexBasedKnife::singleton_ = NULL;

char VertexBasedKnife::str_plugin_title[100];
char VertexBasedKnife::str_command_title[20];

VertexBasedKnife::VertexBasedKnife(void) :
bActive_(FALSE),
bMoved_(FALSE),
bSnaped_(FALSE)
{
	highlightVertex_.object = -1;
	highlightVertex_.vertex = -1;

	baseVertex_.object = -1;
	baseVertex_.vertex = -1;

	endVertex_.object = -1;
	endVertex_.vertex = -1;
}

VertexBasedKnife::VertexBasedKnife( const VertexBasedKnife& p )
{
}

VertexBasedKnife::~VertexBasedKnife()
{
}

VertexBasedKnife *VertexBasedKnife::GetInstance()
{
	if (singleton_ == NULL)
	{
		singleton_ = new VertexBasedKnife();
	}

	return singleton_;
}

void VertexBasedKnife::DeleteInstance()
{
	if (singleton_ != NULL)
	{
		delete singleton_;
	}

	singleton_ = NULL;
}

//---------------------------------------------------------------------------
//  ScissorPlugin::GetPlugInID
//    vOCIDԂB
//---------------------------------------------------------------------------
void VertexBasedKnife::GetPlugInID(DWORD *Product, DWORD *ID)
{
	// v_Ng(Җ)IDAS64bit̒lƂĕԂ
	// l͑ƏdȂ悤ȃ_Ȃ̂ŗǂ
	//*Product = 0xFAC0CF9B;
	//*ID      = 0x7B33E581;
	*Product = 0xE3F7A830;
	*ID = 0xBD9A1882;
}


//---------------------------------------------------------------------------
//  ScissorPlugin::GetPlugInName
//    vOCԂB
//---------------------------------------------------------------------------
const char *VertexBasedKnife::GetPlugInName(void)
{
	std::wstring lang = GetSettingValue(MQSETTINGVALUE_LANGUAGE);
	if (lang == L"Japanese") {
		return "_iCtv4 ver1.1\tCopyright 2012, Cool0707\tCopyright 2014, Tark";
	}
	else {
		return "VertexBaseKnife for v4 ver1.1\tCopyright 2012, Cool0707\tCopyright 2014, Tark";
	}
}


//---------------------------------------------------------------------------
//  ScissorPlugin::EnumString
//    {^ɕ\镶ԂB
//---------------------------------------------------------------------------
const char *VertexBasedKnife::EnumString(void)
{
	std::wstring lang = GetSettingValue(MQSETTINGVALUE_LANGUAGE);
	if (lang == L"Japanese") {
		return "_iCt";
	}
	else {
		return "VertexBaseKnife";
	}
}


//---------------------------------------------------------------------------
//  ScissorPlugin::Initialize
//    AvP[V̏
//---------------------------------------------------------------------------
BOOL VertexBasedKnife::Initialize()
{
	return TRUE;
}

//---------------------------------------------------------------------------
//  ScissorPlugin::Exit
//    AvP[V̏I
//---------------------------------------------------------------------------
void VertexBasedKnife::Exit()
{
	document_ = NULL;
}

void VertexBasedKnife::OnNewDocument(MQDocument doc, const char *filename, NEW_DOCUMENT_PARAM& param)
{
	document_ = doc;
}

//---------------------------------------------------------------------------
//  ScissorPlugin::Activate
//    \E\؂ւv
//---------------------------------------------------------------------------
BOOL VertexBasedKnife::Activate(MQDocument doc, BOOL flag)
{
	bActive_ = (flag == TRUE);

	if(flag)
	{
		SetMouseCursor(GetResourceCursor(MQCURSOR_MOVE));
	}
	else
	{	
		SetMouseCursor(GetResourceCursor(MQCURSOR_DEFAULT));

		highlightVertex_.object = -1;
		highlightVertex_.vertex = -1;

		baseVertex_.object = -1;
		baseVertex_.vertex = -1;

		endVertex_.object = -1;
		endVertex_.vertex = -1;

		bSnaped_ = FALSE;
		bMoved_ = FALSE;

		RedrawAllScene();
	}

	// ̂܂flagԂ
	return flag;
}

//---------------------------------------------------------------------------
//  ScissorPlugin::OnDraw
//    `掞̏
//---------------------------------------------------------------------------
void VertexBasedKnife::OnDraw(MQDocument doc, MQScene scene, int width, int height)
{
	if(bActive_)
	{
		MQObject drawingObject;

		// _nCCg̕`
		if(highlightVertex_.object != -1 && highlightVertex_.vertex != -1) {
			MQObject obj = doc->GetObject(highlightVertex_.object);
			if(obj == NULL) return; // 蓾Ȃ͂AÔNULL`FbN

			// _ʒuvZ
			MQPoint newvp = obj->GetVertex(highlightVertex_.vertex);

			// `IuWFNg𐶐
			int vertex[1];
			drawingObject = CreateDrawingObject(doc, DRAW_OBJECT_POINT);
			drawingObject->SetColor(MQColor(1, 1, 0));
			drawingObject->SetColorValid(TRUE);
			vertex[0] = drawingObject->AddVertex(newvp);
			drawingObject->AddFace(1, vertex);
		}
		// iCtC̕`
		if(baseVertex_.object != -1 && baseVertex_.vertex != -1) {
			// _ʒuvZ
			MQPoint v1, v2;
			v1 = scene->ConvertScreenTo3D(MQPoint(baseScreenPos_.x,
				baseScreenPos_.y,
				scene->GetFrontZ()));
			v2 = scene->ConvertScreenTo3D(MQPoint(endScreenPos_.x,
				endScreenPos_.y,
				scene->GetFrontZ()));

			// `IuWFNg𐶐
			int vertex[2];
			drawingObject = CreateDrawingObject(doc, DRAW_OBJECT_LINE);
			drawingObject->SetColor(MQColor(0, 1, 1));
			drawingObject->SetColorValid(TRUE);
			vertex[0] = drawingObject->AddVertex(v1);
			vertex[1] = drawingObject->AddVertex(v2);
			drawingObject->AddFace(2, vertex);
		}
	}
}

//---------------------------------------------------------------------------
//  VertexBasedKnife::OnLeftButtonDown
//    {^ꂽƂ
//---------------------------------------------------------------------------
BOOL VertexBasedKnife::OnLeftButtonDown(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state)
{	
	// ҏWIvV擾
	EDIT_OPTION option;
	GetEditOption(option);

	MQSelectVertex nearbyVirtex = GetNearbyVertex(scene, state.MousePos, 10, option.CurrentObjectOnly);

	// ߐڒ_炻̒_n_ƂăiCtJn
	if (nearbyVirtex.vertex != -1)
	{
		baseVertex_ = nearbyVirtex;

		baseScreenPos_ = scene->Convert3DToScreen(document_->GetObject(baseVertex_.object)->GetVertex(baseVertex_.vertex));
		endScreenPos_ = baseScreenPos_;

		bSnaped_ = FALSE;

		RedrawScene(scene);
	}

	// W̑ɓƎŝTRUEԂ
	return TRUE;
}

//---------------------------------------------------------------------------
//  VertexBasedKnife::OnLeftButtonUp
//    {^ꂽƂ
//---------------------------------------------------------------------------
BOOL VertexBasedKnife::OnLeftButtonUp(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state)
{
	if (baseVertex_.object != -1 && baseVertex_.vertex != -1 && bMoved_)
	{
		if(!state.RButton && !state.Alt) {
			// {^ꂽƂɁAALTE{^ĂȂƂ̓XibvȂ
			bSnaped_ = FALSE;
			endVertex_.object = -1;
			endVertex_.vertex = -1;
		}

		MQPoint tsv1, tsv2;

		tsv1.x = (float)baseScreenPos_.x;
		tsv1.y = (float)baseScreenPos_.y;
		tsv1.z = 0.0f;
		tsv2.x = (float)endScreenPos_.x;
		tsv2.y = (float)endScreenPos_.y;
		tsv2.z = 0.0f;

		// iCtŐؒf
		if (ExecKnifeCut(doc, scene, tsv1, tsv2, baseVertex_, endVertex_))
		{
			// ؒfꍇAĕ`ƃAhD̍XVs
			RedrawAllScene();
			UpdateUndo();
		}
	}

	baseVertex_.object = -1;
	baseVertex_.vertex = -1;
	bMoved_ = FALSE;

	// W̑ɓƎŝTRUEԂ
	return TRUE;
}

//---------------------------------------------------------------------------
//  VertexBasedKnife::OnLeftButtonMove
//    {^Ȃ}EXړƂ
//---------------------------------------------------------------------------
BOOL VertexBasedKnife::OnLeftButtonMove(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state)
{
	if (state.RButton || state.Alt)
	{
		if (baseVertex_.object != -1 && baseVertex_.vertex != -1)
		{
			// ҏWIvV擾
			EDIT_OPTION option;
			GetEditOption(option);

			MQSelectVertex nearbyVirtex = GetNearbyVertex(scene, state.MousePos, 10, option.CurrentObjectOnly);

			if (nearbyVirtex.vertex != -1)
			{
				endScreenPos_ = scene->Convert3DToScreen(doc->GetObject(nearbyVirtex.object)->GetVertex(nearbyVirtex.vertex));
				bSnaped_ = TRUE;
				endVertex_ = nearbyVirtex;

				RedrawScene(scene);
				bMoved_ = TRUE;
			}

			return TRUE;
		}
	}
	if (baseVertex_.object != -1 && baseVertex_.vertex != -1)
	{
		// n_ƃJ[\ʒu
		if (state.Shift)
		{
			if (abs(state.MousePos.x-baseScreenPos_.x) >= abs(state.MousePos.y-baseScreenPos_.y))
			{
				endScreenPos_.x = (float)state.MousePos.x;
				endScreenPos_.y = baseScreenPos_.y;
			}
			else
			{
				endScreenPos_.x = baseScreenPos_.x;
				endScreenPos_.y = (float)state.MousePos.y;
			}
		}
		else
		{
			endScreenPos_.x = (float)state.MousePos.x;
			endScreenPos_.y = (float)state.MousePos.y;
		}

		// ĕ`
		RedrawScene(scene);

		// ړς݃`FbN
		bMoved_ = TRUE;

	}

	// W̑ɓƎŝTRUEԂ
	return TRUE;
}

//---------------------------------------------------------------------------
//  VertexBasedKnife::OnMouseMove
//    }EXړƂ
//---------------------------------------------------------------------------
BOOL VertexBasedKnife::OnMouseMove(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state)
{
	MQSelectVertex newVirtex;

	// ҏWIvV擾
	EDIT_OPTION option;
	GetEditOption(option);

	newVirtex = GetNearbyVertex(scene, state.MousePos, 10, option.CurrentObjectOnly);

	// OƈႤ_ɃJ[\ꍇ͍ĕ`
	if (newVirtex != highlightVertex_)
	{
		highlightVertex_ = newVirtex;
		RedrawScene(scene);
	}

	// ƎsAWs킹邽FALSEԂ
	return FALSE;
}

/**
* XN[Wŋߐڒ_擾.
* 
* @param scene ΏۃXN[
* @param screenPos eXgW
* @param range eXg͈
* @param currentObjectOnly ݃IuWFNĝ݂ǂ
*
* @return ŋߐڒ_
*/
MQSelectVertex VertexBasedKnife::GetNearbyVertex(MQScene scene, POINT screenPos, float range, BOOL currentObjectOnly)
{
	MQSelectVertex nearbyVertex;
	float minDist = range * range;

	nearbyVertex.object = -1;
	nearbyVertex.vertex = -1;

	int currentObjectIndex = document_->GetCurrentObjectIndex();

	int numObj = document_->GetObjectCount();
	for (int oi=0; oi<numObj; oi++)
	{
		MQObject obj = document_->GetObject(oi);
		if (obj == NULL) continue;

		// \܂̓bNĂΑΏۊO
		if (!obj->GetVisible() || obj->GetLocking()) continue;

		// JgIuWFNĝ
		if (currentObjectOnly && oi != currentObjectIndex)
			continue;

		int numVertex = obj->GetVertexCount();
		for (int vi=0; vi<numVertex; vi++)
		{
			if (obj->GetVertexRefCount(vi) != 0 && (baseVertex_.object != oi || baseVertex_.vertex != vi))
			{
				// NbNꂽʒuɋ߂ǂ
				MQPoint sp = scene->Convert3DToScreen(obj->GetVertex(vi));
				float d = (sp.x-screenPos.x)*(sp.x-screenPos.x) + (sp.y-screenPos.y)*(sp.y-screenPos.y);
				if (minDist > d + FLT_MIN)
				{
					minDist = d;
					nearbyVertex.object = oi;
					nearbyVertex.vertex = vi;

				}
				if(currentObjectIndex == oi && std::fabs(minDist - d) <= FLT_MIN)
				{
					minDist = d;
					nearbyVertex.object = oi;
					nearbyVertex.vertex = vi;
				}
			}
		}
	}

	return nearbyVertex;
}

struct FaceStruct
{
	std::vector<int> varray;
	std::vector<MQCoordinate> carray;
	std::vector<DWORD> colarray;
};

struct KnifeHash {
	KnifeHash *next;
	int v2;
	int nvi; // VKɏo_̃CfbNX
	float t;
	MQPoint v;
};

class KnifeHashList
{
private:
	KnifeHash *hash_;
	int size_;

public:
	KnifeHashList(int maxhSize, int size)
	{
		hash_ = new KnifeHash[maxhSize];

		for(int i=0; i<maxhSize; i++)
		{
			hash_[i].next = NULL;
			hash_[i].v2 = hash_[i].nvi = -1;
		}

		size_ = size;
	};

	~KnifeHashList()
	{
		delete[] hash_;
	};

	KnifeHash *Begin()
	{
		return hash_;
	};

	KnifeHash *End()
	{
		return &hash_[size_];
	};

	int Size()
	{
		return size_;
	};

	bool IsSet(int v1, int v2)
	{
		for(KnifeHash *ch=&hash_[v1]; ch!=NULL && ch->v2>=0; ch=ch->next)
		{
			if(ch->v2 == v2)
			{
				return true;
			}
		}

		return false;
	};

	KnifeHash Find(int v1, int v2)
	{
		KnifeHash dmy;

		dmy.next = NULL;
		dmy.nvi = -1;
		
		for(KnifeHash *ch=&hash_[v1]; ch!=NULL && ch->v2>=0; ch=ch->next)
		{
			if(ch->v2 == v2)
			{
				return *ch;
			}
		}

		return dmy;
	};

	void Add(MQScene scene, MQObject obj, int vi1, int vi2, float t)
	{
		float w1, w2;
		MQPoint sv1 = scene->Convert3DToScreen(obj->GetVertex(vi1), &w1);
		MQPoint sv2 = scene->Convert3DToScreen(obj->GetVertex(vi2), &w2);

		for(KnifeHash *ch=&hash_[vi1]; ; ch=ch->next)
		{
			if (ch->v2 < 0)
			{
				ch->v2 = vi2;

				MQPoint v0 = scene->ConvertScreenTo3D(sv1*(1-t) + sv2*t);
				MQPoint v1 = obj->GetVertex(vi1);
				MQPoint v2 = obj->GetVertex(vi2);

				float d1 = GetSize(v0 - v1);
				float d2 = GetSize(v2 - v1);

				if(d2 > 1e-8f)
					ch->t = d1 / d2;
				else
					ch->t = t;

				ch->v = v1 * (1-ch->t) + v2 * ch->t;
				break;
			}

			if(ch->next == NULL)
				ch->next = &hash_[size_++];
		}
	};
};

KnifeHashList *GetCrossingLines(MQObject obj, MQScene scene, MQPoint tsv1, MQPoint tsv2, int kvi1, int kvi2)
{
	// ʐ
	int faceCount = obj->GetFaceCount();
	// _
	int vertCount = obj->GetVertexCount();

	// ʂ邩ǂׂĂ
	BOOL *visible = new BOOL[faceCount];
	scene->GetVisibleFace(obj, visible);

	// allocate hash
	KnifeHashList *hashList = new KnifeHashList(faceCount*4 + vertCount, vertCount);

	// Ă郉Cm
	int count = 0;
	for(int fi=0; fi<faceCount; fi++)
	{
		int facePointCount = obj->GetFacePointCount(fi);

		// _܂܂Ȃʂ͖
		if(facePointCount == 0) continue;

		// s̖ʂ͖
		if(!visible[fi]) continue;

		std::vector<int> vert(facePointCount);
		obj->GetFacePointArray(fi, &vert[0]);

		for(int vi=0; vi<facePointCount; vi++)
		{
			// ̏I_Ȃ烋[vo
			if(vi == 1 && facePointCount == 2)
				break;

			// check hash
			int hv1 = min(vert[vi], vert[(vi+1) % facePointCount]);
			int hv2 = max(vert[vi], vert[(vi+1) % facePointCount]);

			if (hashList->IsSet(hv1, hv2))
				continue;

			// check if crossing
			MQPoint screen1,screen2;
			float w1,w2;
			screen1 = scene->Convert3DToScreen(obj->GetVertex(hv1),&w1);
			screen2 = scene->Convert3DToScreen(obj->GetVertex(hv2),&w2);
			float t;
			if(!IsCrossing2D(tsv1, tsv2, screen1, screen2, NULL, &t))
				continue;

			if(kvi1 == hv1 || kvi1 == hv2)
				continue;

			if(kvi2 == hv1 || kvi2 == hv2)
				continue;

			// add to hash
			hashList->Add(scene, obj, hv1, hv2, t);

			count++;
		}
	}

	delete[] visible;

	// // ؒf郉CȂNULLԂ
	if( count == 0)
	{
		delete hashList;
		return NULL;
	}

	return hashList;
}

void AddSegment(MQObject obj, int vert0, int vert1, int mat)
{
	int vert[2];
	vert[0] = vert0;
	vert[1] = vert1;

	int face = obj->AddFace(2, vert);
	
	obj->SetFaceMaterial(face, mat);
}

MQSelectFace VertexBasedKnife::GetCommonFace(MQSelectVertex v1, MQSelectVertex v2)
{
	MQSelectFace commonFace;

	commonFace.object = -1;
	commonFace.face = -1;
	
	if (v1.object != v2.object)
	{
		return commonFace;
	}

	MQObject obj = document_->GetObject(v1.object);

	if (obj == NULL)
	{
		return commonFace;
	}

	for (int fi=0; fi<obj->GetFaceCount(); fi++)
	{
		int facePointCount = obj->GetFacePointCount(fi);
		if(facePointCount == 0) {
			continue;
		}
		std::vector<int> vertexIndexes(facePointCount);
		obj->GetFacePointArray(fi, &vertexIndexes[0]);

		bool v1Flag = false;
		bool v2Flag = false;

		for(int fvi=0; fvi<obj->GetFacePointCount(fi); fvi++)
		{
			if (vertexIndexes[fvi] == v1.vertex)
			{
				v1Flag = true;
			}
			if (vertexIndexes[fvi] == v2.vertex)
			{
				v2Flag = true;
			}
		}

		if (v1Flag && v2Flag)
		{
			commonFace.object = v1.object;
			commonFace.face = fi;
		}
	}

	return commonFace;
}

void AddTriangle(MQObject obj,
				 int vert0, int vert1, int vert2,
				 int mat,
				 MQCoordinate coord0, MQCoordinate coord1, MQCoordinate coord2,
				 DWORD color0, DWORD color1, DWORD color2)
{
	int vert[3];
	vert[0] = vert0;
	vert[1] = vert1;
	vert[2] = vert2;
	int face = obj->AddFace(3, vert);
	
	obj->SetFaceMaterial(face, mat);
	
	MQCoordinate coord[3];
	coord[0] = coord0;
	coord[1] = coord1;
	coord[2] = coord2;
	obj->SetFaceCoordinateArray(face, coord);

	obj->SetFaceVertexColor(face, 0, color0);
	obj->SetFaceVertexColor(face, 1, color1);
	obj->SetFaceVertexColor(face, 2, color2);
}

void AddSquare(MQObject obj,
				 int vert0, int vert1, int vert2, int vert3,
				 int mat,
				 MQCoordinate coord0, MQCoordinate coord1, MQCoordinate coord2, MQCoordinate coord3,
				 DWORD color0, DWORD color1, DWORD color2, DWORD color3)
{
	int vert[4];
	vert[0] = vert0;
	vert[1] = vert1;
	vert[2] = vert2;
	vert[3] = vert3;
	int face = obj->AddFace(4, vert);
	
	obj->SetFaceMaterial(face, mat);
	
	MQCoordinate coord[4];
	coord[0] = coord0;
	coord[1] = coord1;
	coord[2] = coord2;
	coord[3] = coord3;
	obj->SetFaceCoordinateArray(face, coord);

	obj->SetFaceVertexColor(face, 0, color0);
	obj->SetFaceVertexColor(face, 1, color1);
	obj->SetFaceVertexColor(face, 2, color2);
	obj->SetFaceVertexColor(face, 3, color3);
}

void AddPolygon(MQObject obj, std::vector<int> &nfvi, int mat, std::vector<MQCoordinate> &ncoord, std::vector<DWORD> &ncolor)
{
	int face = obj->AddFace(nfvi.size(), &nfvi[0]);
	obj->SetFaceMaterial(face, mat);
	obj->SetFaceCoordinateArray(face, &ncoord[0]);
	for(int i = 0; i<ncolor.size(); i++) {
		obj->SetFaceVertexColor(face, i, ncolor[i]);
	}
}

void AddPolygonbyVindex(MQObject obj, const std::vector<int> &nidx,
	const std::vector<int> &fvi, int mat, const std::vector<MQCoordinate> &coord, const std::vector<DWORD> &color)
{
	std::vector<int> nfvi(nidx.size());
	std::vector<MQCoordinate> ncoord(nidx.size());
	std::vector<DWORD> ncolor(nidx.size());

	for(int i=0; i<nidx.size(); i++)
	{
		nfvi[i] = fvi[nidx[i]];
		ncoord[i] = coord[nidx[i]];
		ncolor[i] = color[nidx[i]];
	}
	AddPolygon(obj, nfvi, mat, ncoord, ncolor);
}

void CutByHashList(MQObject obj, MQScene scene, KnifeHashList *hashList, int kvi1, int kvi2)
{
	// ʐ
	int faceCount = obj->GetFaceCount();
	// _
	int vertCount = obj->GetVertexCount();

	// ؒfC̒_ǉ
	for (KnifeHash *ch=hashList->Begin(); ch!=hashList->End(); ch++)
	{
		if(ch->v2 >= 0)
		{
			ch->nvi = obj->AddVertex(ch->v);
		}		
	}

	// ؒfʂ̍\z
	for (int fi=0; fi<faceCount; fi++)
	{
		int faceVertCount = obj->GetFacePointCount(fi);
		if(obj->GetFacePointCount(fi) == 0) continue;

		std::vector<int> nvi(faceVertCount);
		std::vector<MQCoordinate> nfcoord(faceVertCount);
		std::vector<float> t(faceVertCount);
		std::vector<int> cfvi(faceVertCount);
		std::vector<MQCoordinate> cfcoord(faceVertCount);
		std::vector<DWORD> cfcolor(faceVertCount);

		// _CfbNXz̎擾
		obj->GetFacePointArray(fi, &cfvi[0]);
		// UVz̎擾
		obj->GetFaceCoordinateArray(fi, &cfcoord[0]);
		// _J[z̎擾
		for (int i=0; i<faceVertCount; i++)
		{
			cfcolor[i] = obj->GetFaceVertexColor(fi, i);
		}
		// ގ̎擾
		int pal = obj->GetFaceMaterial(fi);

		// f郉C̒_CfbNXnviɃZbg
		for (int i=0; i<faceVertCount; i++)
		{
			KnifeHash h = hashList->Find(
				min(cfvi[i], cfvi[(i+1)%faceVertCount]), 
				max(cfvi[i], cfvi[(i+1)%faceVertCount]));
			nvi[i] = h.nvi;
			t[i] = h.t;
		}

		int kv = -1;
		for (int i=0; i<faceVertCount; i++)
		{
			if(kvi1 == cfvi[i])
			{
				kv = i;
				break;
			}
			if(kvi2 == cfvi[i])
			{
				kv = i;
				break;
			}
		}

		// ʁECؒf
		switch(faceVertCount)
		{
		case 2:
			if (nvi[0]>=0)
			{
				AddSegment(obj, cfvi[0], nvi[0], pal);
				
				AddSegment(obj, cfvi[1], nvi[0], pal);

				obj->DeleteFace(fi);
			}
			break;

		case 3:
			for (int k=0; k<3; k++)
			{
				if (nvi[k]>=0 && nvi[(k+1)%3]>=0)
				{
					MQCoordinate t0,t1;
					DWORD col0,col1;
					if(cfvi[k] < cfvi[(k+1)%3]){
						t0 = cfcoord[k]*(1-t[k]) + cfcoord[(k+1)%3]*t[k];
						col0 = (MQRGBAColor(cfcolor[k])*(1-t[k]) + MQRGBAColor(cfcolor[(k+1)%3])*t[k]).GetRGBA();
					}else{
						t0 = cfcoord[k]*t[k] + cfcoord[(k+1)%3]*(1-t[k]);
						col0 = (MQRGBAColor(cfcolor[k])*t[k] + MQRGBAColor(cfcolor[(k+1)%3])*(1-t[k])).GetRGBA();
					}
					if(cfvi[(k+1)%3] < cfvi[(k+2)%3]){
						t1 = cfcoord[(k+1)%3]*(1-t[(k+1)%3]) + cfcoord[(k+2)%3]*t[(k+1)%3];
						col1 = (MQRGBAColor(cfcolor[(k+1)%3])*(1-t[(k+1)%3]) + MQRGBAColor(cfcolor[(k+2)%3])*t[(k+1)%3]).GetRGBA();
					}else{
						t1 = cfcoord[(k+1)%3]*t[(k+1)%3] + cfcoord[(k+2)%3]*(1-t[(k+1)%3]);
						col1 = (MQRGBAColor(cfcolor[(k+1)%3])*t[(k+1)%3] + MQRGBAColor(cfcolor[(k+2)%3])*(1-t[(k+1)%3])).GetRGBA();
					}

					AddTriangle(obj, nvi[k], cfvi[(k+1)%3], nvi[(k+1)%3], pal, t0, cfcoord[(k+1)%3], t1, col0, cfcolor[(k+1)%3], col1);
					
					AddSquare(obj, nvi[k], nvi[(k+1)%3], cfvi[(k+2)%3], cfvi[k], pal,
						t0, t1, cfcoord[(k+2)%3], cfcoord[k],
						col0, col1, cfcolor[(k+2)%3], cfcolor[k]);

					obj->DeleteFace(fi);
					break;
				}
				if (nvi[k]>=0 && kv>=0)
				{
					MQCoordinate t0;
					DWORD col0;
					if(cfvi[k] < cfvi[(k+1)%3]){
						t0 = cfcoord[k]*(1-t[k]) + cfcoord[(k+1)%3]*t[k];
						col0 = (MQRGBAColor(cfcolor[k])*(1-t[k]) + MQRGBAColor(cfcolor[(k+1)%3])*t[k]).GetRGBA();
					}else{
						t0 = cfcoord[k]*t[k] + cfcoord[(k+1)%3]*(1-t[k]);
						col0 = (MQRGBAColor(cfcolor[k])*t[k] + MQRGBAColor(cfcolor[(k+1)%3])*(1-t[k])).GetRGBA();
					}

					AddTriangle(obj, nvi[k], cfvi[kv], cfvi[(kv+1)%3], pal, t0, cfcoord[kv], cfcoord[(kv+1)%3], col0, cfcolor[kv], cfcolor[(kv+1)%3]);
					AddTriangle(obj, nvi[k], cfvi[(kv+2)%3], cfvi[kv], pal, t0, cfcoord[kv], cfcoord[(kv+1)%3], col0, cfcolor[kv], cfcolor[(kv+1)%3]);

					obj->DeleteFace(fi);
					break;
				}
			}
			break;

		case 4:
			for(int k=0; k<4; k++)
			{
				if(nvi[k]>=0 && nvi[(k+2)%4]>=0)
				{
					MQCoordinate t0,t1;
					DWORD col0,col1;
					if(cfvi[k] < cfvi[(k+1)%4]){
						t0 = cfcoord[k]*(1-t[k]) + cfcoord[(k+1)%4]*t[k];
						col0 = (MQRGBAColor(cfcolor[k])*(1-t[k]) + MQRGBAColor(cfcolor[(k+1)%4])*t[k]).GetRGBA();
					}else{
						t0 = cfcoord[k]*t[k] + cfcoord[(k+1)%4]*(1-t[k]);
						col0 = (MQRGBAColor(cfcolor[k])*t[k] + MQRGBAColor(cfcolor[(k+1)%4])*(1-t[k])).GetRGBA();
					}
					if(cfvi[(k+2)%4] < cfvi[(k+3)%4]){
						t1 = cfcoord[(k+2)%4]*(1-t[(k+2)%4]) + cfcoord[(k+3)%4]*t[(k+2)%4];
						col1 = (MQRGBAColor(cfcolor[(k+2)%4])*(1-t[(k+2)%4]) + MQRGBAColor(cfcolor[(k+3)%4])*t[(k+2)%4]).GetRGBA();
					}else{
						t1 = cfcoord[(k+2)%4]*t[(k+2)%4] + cfcoord[(k+3)%4]*(1-t[(k+2)%4]);
						col1 = (MQRGBAColor(cfcolor[(k+2)%4])*t[(k+2)%4] + MQRGBAColor(cfcolor[(k+3)%4])*(1-t[(k+2)%4])).GetRGBA();
					}

					AddSquare(obj, nvi[k], cfvi[(k+1)%4], cfvi[(k+2)%4], nvi[(k+2)%4], pal,
						t0, cfcoord[(k+1)%4], cfcoord[(k+2)%4], t1,
						col0, cfcolor[(k+1)%4], cfcolor[(k+2)%4], col1);

					AddSquare(obj, nvi[k], nvi[(k+2)%4], cfvi[(k+3)%4], cfvi[k], pal,
						t0, t1, cfcoord[(k+3)%4], cfcoord[k],
						col0, col1, cfcolor[(k+3)%4], cfcolor[k]);

					obj->DeleteFace(fi);
					break;
				}
				if(nvi[k]>=0 && nvi[(k+1)%4]>=0)
				{
					MQCoordinate t0,t1;
					DWORD col0,col1;
					if(cfvi[k] < cfvi[(k+1)%4]){
						t0 = cfcoord[k]*(1-t[k]) + cfcoord[(k+1)%4]*t[k];
						col0 = (MQRGBAColor(cfcolor[k])*(1-t[k]) + MQRGBAColor(cfcolor[(k+1)%4])*t[k]).GetRGBA();
					}else{
						t0 = cfcoord[k]*t[k] + cfcoord[(k+1)%4]*(1-t[k]);
						col0 = (MQRGBAColor(cfcolor[k])*t[k] + MQRGBAColor(cfcolor[(k+1)%4])*(1-t[k])).GetRGBA();
					}
					if(cfvi[(k+1)%4] < cfvi[(k+2)%4]){
						t1 = cfcoord[(k+1)%4]*(1-t[(k+1)%4]) + cfcoord[(k+2)%4]*t[(k+1)%4];
						col1 = (MQRGBAColor(cfcolor[(k+1)%4])*(1-t[(k+1)%4]) + MQRGBAColor(cfcolor[(k+2)%4])*t[(k+1)%4]).GetRGBA();
					}else{
						t1 = cfcoord[(k+1)%4]*t[(k+1)%4] + cfcoord[(k+2)%4]*(1-t[(k+1)%4]);
						col1 = (MQRGBAColor(cfcolor[(k+1)%4])*t[(k+1)%4] + MQRGBAColor(cfcolor[(k+2)%4])*(1-t[(k+1)%4])).GetRGBA();
					}

					AddTriangle(obj, nvi[k], cfvi[(k+1)%4], nvi[(k+1)%4], pal, t0, cfcoord[(k+1)%4], t1, col0,  cfcolor[(k+1)%4], col1);
					
					AddSquare(obj, nvi[k], nvi[(k+1)%4], cfvi[(k+2)%4], cfvi[(k+3)%4], pal,
						t0, t1, cfcoord[(k+2)%4], cfcoord[(k+3)%4],
						col0, col1, cfcolor[(k+2)%4], cfcolor[(k+3)%4]);

					AddTriangle(obj, nvi[k], cfvi[(k+3)%4], cfvi[k], pal, t0, cfcoord[(k+3)%4], cfcoord[k], col0,  cfcolor[(k+3)%4], cfcolor[k]);

					obj->DeleteFace(fi);
					break;
				}
				if (nvi[k]>=0 && kv == (k+2)%4)
				{
					MQCoordinate t0;
					DWORD col0;
					if(cfvi[k] < cfvi[(k+1)%4]){
						t0 = cfcoord[k]*(1-t[k]) + cfcoord[(k+1)%4]*t[k];
						col0 = (MQRGBAColor(cfcolor[k])*(1-t[k]) + MQRGBAColor(cfcolor[(k+1)%4])*t[k]).GetRGBA();
					}else{
						t0 = cfcoord[k]*t[k] + cfcoord[(k+1)%4]*(1-t[k]);
						col0 = (MQRGBAColor(cfcolor[k])*t[k] + MQRGBAColor(cfcolor[(k+1)%4])*(1-t[k])).GetRGBA();
					}

					AddTriangle(obj, nvi[k], cfvi[(kv+3)%4], cfvi[kv], pal, t0, cfcoord[(kv+3)%4], cfcoord[kv], col0, cfcolor[(kv+3)%4], cfcolor[kv]);
					
					AddSquare(obj, nvi[k], cfvi[kv], cfvi[(kv+1)%4], cfvi[(kv+2)%4], pal,
						t0, cfcoord[kv], cfcoord[(kv+1)%4], cfcoord[(kv+2)%4],
						col0, cfcolor[kv], cfcolor[(kv+1)%4], cfcolor[(kv+2)%4]);

					obj->DeleteFace(fi);
					break;
				}
				if (nvi[k]>=0 && kv == (k+3)%4)
				{
					MQCoordinate t0;
					DWORD col0;
					if(cfvi[k] < cfvi[(k+1)%4]){
						t0 = cfcoord[k]*(1-t[k]) + cfcoord[(k+1)%4]*t[k];
						col0 = (MQRGBAColor(cfcolor[k])*(1-t[k]) + MQRGBAColor(cfcolor[(k+1)%4])*t[k]).GetRGBA();
					}else{
						t0 = cfcoord[k]*t[k] + cfcoord[(k+1)%4]*(1-t[k]);
						col0 = (MQRGBAColor(cfcolor[k])*t[k] + MQRGBAColor(cfcolor[(k+1)%4])*(1-t[k])).GetRGBA();
					}

					AddTriangle(obj, nvi[k], cfvi[kv], cfvi[(kv+1)%4], pal, t0, cfcoord[kv], cfcoord[(kv+1)%4], col0, cfcolor[kv], cfcolor[(kv+1)%4]);
					
					AddSquare(obj, nvi[k], cfvi[(kv+2)%4], cfvi[(kv+3)%4], cfvi[kv], pal,
						t0, cfcoord[(kv+2)%4], cfcoord[(kv+3)%4], cfcoord[kv],
						col0, cfcolor[(kv+2)%4], cfcolor[(kv+3)%4], cfcolor[kv]);

					obj->DeleteFace(fi);
					break;
				}
			}
			break;
		default: // p`
			{
				// ʂ̍\ӂāACOŐ؂ւāAXgɑ}ĂƂ
				// 2ʂ̒_/UV/J[̃Xg쐬
				int phase = 0;
				FaceStruct fs[2];
				for(int k = 0; k < faceVertCount; k++) {
					if(kv == k) {
						fs[phase].varray.push_back(cfvi[k]);
						fs[phase].carray.push_back(cfcoord[k]);
						fs[phase].colarray.push_back(cfcolor[k]);
						phase = 1 - phase;
					}
					fs[phase].varray.push_back(cfvi[k]);
					fs[phase].carray.push_back(cfcoord[k]);
					fs[phase].colarray.push_back(cfcolor[k]);
					if(nvi[k] >= 0) {
						MQCoordinate t1;
						DWORD col1;
						if(cfvi[k] < cfvi[(k + 1) % faceVertCount]) {
							t1 = cfcoord[k] * (1 - t[k]) + cfcoord[(k + 1) % faceVertCount] * t[k];
							col1 = (MQRGBAColor(cfcolor[k])*(1 - t[k]) + MQRGBAColor(cfcolor[(k + 1) % faceVertCount])*t[k]).GetRGBA();
						}
						else {
							t1 = cfcoord[k] * t[k] + cfcoord[(k + 1) % faceVertCount] * (1 - t[k]);
							col1 = (MQRGBAColor(cfcolor[k])*t[k] + MQRGBAColor(cfcolor[(k + 1) % faceVertCount])*(1 - t[k])).GetRGBA();
						}
						fs[phase].varray.push_back(nvi[k]);
						fs[phase].carray.push_back(t1);
						fs[phase].colarray.push_back(col1);
						phase = 1 - phase;
						fs[phase].varray.push_back(nvi[k]);
						fs[phase].carray.push_back(t1);
						fs[phase].colarray.push_back(col1);
					}
				}
				// 쐬Xg2ʂ𐶐
				for(int i = 0; i < 2; i++) {
					AddPolygon(obj, fs[i].varray, pal, fs[i].carray, fs[i].colarray);
				}
				obj->DeleteFace(fi);
			}
		}
	}
}

/**
* ؒf̎s
*/
bool VertexBasedKnife::ExecKnifeCut(MQDocument doc, MQScene scene, MQPoint tsv1, MQPoint tsv2, MQSelectVertex kv1, MQSelectVertex kv2)
{
	// ҏWIvV擾
	EDIT_OPTION option;
	GetEditOption(option);

	// ؒfʂ̖@߂
	MQPoint plane1 = scene->ConvertScreenTo3D(tsv1);
	MQPoint plane2 = scene->ConvertScreenTo3D(tsv2);
	MQPoint plane3 = scene->ConvertScreenTo3D(MQPoint(tsv1.x,tsv1.y,0.25f));
	MQPoint normal = GetNormal(plane1, plane2, plane3);

	MQSelectFace commonFace = GetCommonFace(kv1, kv2);

	if (commonFace.object != -1&& commonFace.face != -1)
	{	
		MQObject obj = doc->GetObject(commonFace.object);

		if (obj == NULL) return false;

		// \܂̓bNĂΑΏۊO
		if (!obj->GetVisible() || obj->GetLocking()) return false;
		
		// JgIuWFNĝݕҏW
		if (option.CurrentObjectOnly && commonFace.object != doc->GetCurrentObjectIndex()) return false;

		// lpȊO͐ؒfs\
		int facePointCount = obj->GetFacePointCount(commonFace.face);
		if (facePointCount < 4) return false;
		
		std::vector<int> fvi(facePointCount);
		obj->GetFacePointArray(commonFace.face, &fvi[0]);

		std::vector<MQCoordinate> coord(facePointCount);
		// UVz̎擾
		obj->GetFaceCoordinateArray(commonFace.face, &coord[0]);
		std::vector<DWORD> color(facePointCount);
		// _J[z̎擾
		for (int i=0; i<obj->GetFacePointCount(commonFace.face); i++)
		{
			color[i] = obj->GetFaceVertexColor(commonFace.face, i);
		}
		// ގ̎擾
		int mat = obj->GetFaceMaterial(commonFace.face);

		std::vector<int> vecindex[2];
		int phase = 0;
		for(int i = 0; i < facePointCount; i++) {
			vecindex[phase].push_back(i);
			if(fvi[i] == kv1.vertex || fvi[i] == kv2.vertex) {
				phase = 1 - phase;
				vecindex[phase].push_back(i);
			}
		}
		if(vecindex[0].size() == 2 || vecindex[1].size() == 2) {
			// אڒ_Őؒf悤Ƃ
			return false;
		}
		for(int i = 0; i < 2; i++) {
			AddPolygonbyVindex(obj, vecindex[i], fvi, mat, coord, color);
		}
		obj->DeleteFace(commonFace.face);
		/*
		for(int i = 0; i<facePointCount; i++)
		{
			if(fvi[i]==kv1.vertex && fvi[(i+2)%4]==kv2.vertex)
			{
				AddTriangle(obj, fvi[i], fvi[(i+1)%4], fvi[(i+2)%4], mat,
					coord[i], coord[(i+1)%4], coord[(i+2)%4],
					color[i], color[(i+1)%4], color[(i+2)%4]);

				AddTriangle(obj, fvi[(i+2)%4], fvi[(i+3)%4], fvi[i], mat,
					coord[(i+2)%4], coord[(i+3)%4], coord[i],
					color[(i+2)%4], color[(i+3)%4], color[i]);

				obj->DeleteFace(commonFace.face);
				break;
			}
		}
*/

		return true;
	}

	int count = 0;

	for(int oi=0; oi<doc->GetObjectCount(); oi++)
	{
		MQObject obj = doc->GetObject(oi);
		if (obj == NULL) continue;

		// \܂̓bNĂΑΏۊO
		if (!obj->GetVisible() || obj->GetLocking()) continue;

		// JgIuWFNĝݕҏW
		if (option.CurrentObjectOnly && oi != doc->GetCurrentObjectIndex()) continue;

		// JnEI_̐ݒ
		int kvi1, kvi2; 
		if (kv1.object == oi) kvi1 = kv1.vertex;
		else kvi1 = -1;
		if (kv2.object == oi) kvi2 = kv2.vertex;
		else kvi2 = -1;

		// Ă郉C̃nbVXg擾
		KnifeHashList *hashList = GetCrossingLines(obj, scene, tsv1, tsv2, kvi1, kvi2);

		// ؒf郉CȂ΃pX
		if (hashList == NULL) continue;

		// nbVXgɐؒf
		CutByHashList(obj, scene, hashList, kvi1, kvi2);

		count++;

		delete hashList;
	}

	return count>0;
}

//---------------------------------------------------------------------------
//  GetPluginClass
//    vOC̃x[XNXԂ
//---------------------------------------------------------------------------
MQBasePlugin *GetPluginClass()
{
	return VertexBasedKnife::GetInstance();
}